
#include <regex>
#include <sdktools>
#include <outputactions>


char listfieldtypes[][] = {
	"Unsupported",      //The type is unsupported. 
	"Integer",          //Valid for SendProp and Data fields 
	"Float",            //Valid for SendProp and Data fields 
	"Entity",           //Valid for Data fields only (SendProp shows as int) 
	"Vector",           //Valid for SendProp and Data fields 
	"String",           //Valid for SendProp and Data fields 
	"String Read Only",         //Valid for Data fields.  Read only.
	                              //Note that the size of a string_t is dynamic, and
	                              //thus FindDataMapOffs() will return the constant size
	                              //of the string_t container (which is 32 bits right now). 
	"Variant"           //Valid for Data fields only Type is not known at the field level,
                        //   (for this call), but dependent on current field value.
}

Regex regexclassname;
Regex regexoutput;
Regex regexinput;
Regex regexkey;

KeyValues ListClassnameOutputs;

public void OnPluginStart()
{
	RegConsoleCmd("sm_lookoutputs", lookoutputs);

	ListClassnameOutputs = new KeyValues("ListClassnameOutputs");

	// If this file exist already, don't read Prop_Data.txt file and create output file.
	if(FileExists("ListClassnameOutputs.kv"))
	{
		if(!ListClassnameOutputs.ImportFromFile("ListClassnameOutputs.kv"))
		{
			SetFailState("Failed import KeyValue file ListClassnameOutputs.kv");
		}

		return;
	}


	LogAction(-1, -1, "Plugin did not find file ListClassnameOutputs.kv. Trying to next, read file Prop_Data.txt file");

	File prop_data_file = OpenFile("Prop_Data.txt", "rt");

	if(prop_data_file == null)
	{
		SetFailState("Plugin couldn't find/open/read Prop_Data.txt file. Plugin now stop.");
	}

	regexclassname = new Regex("(^\\w+) - (\\w+)$");
	regexoutput = new Regex("^\\W+(\\w+).*\\WOutput\\W.*?(\\w+)$");
	regexinput = new Regex("^\\W+(\\w+).*\\WInput\\W.*?(\\w+)$");
	regexkey = new Regex("^\\W+(\\w+).*\\WKey\\W.*?(\\w+)$");


	LogAction(-1, -1, "Plugin found Prop_Data.txt file. Plugin is going to read and create output file ListClassnameOutputs.kv. This can take LONG TIME! Even over 10 minutes if Prop_Data.txt file is big!");

	CreateTimer(5.0, timerdelay, prop_data_file);
}

int g_timestamp;

public Action timerdelay(Handle timer, File prop_data_file)
{

	g_timestamp = GetTime();

	RequestFrame(nextframe, prop_data_file);

	return Plugin_Continue;
}



public void nextframe(File prop_data_file)
{
	static bool IsClassnameEntry = false;
	static bool IsEnd = false;
	static char NetClass[256];
	static char buffer[256];
	static char skeyindex[100];
	static int counter = 0;
	static PropFieldType fieldtype;

	for(int a = 0; a <= 1000; a++)
	{
		IsEnd = prop_data_file.EndOfFile();

		// we are end of file
		if(IsEnd)
		{
			NoteTimePassed(GetTime() - g_timestamp, counter);
			LogAction(-1, -1, "Plugin finished reading file Prop_Data.txt. Saving now file ListClassnameOutputs.kv");

			ListClassnameOutputs.Rewind();
			ListClassnameOutputs.ExportToFile("ListClassnameOutputs.kv");

			delete prop_data_file;
			//return;
			break;
		}

		counter++;

		if(prop_data_file.ReadLine(buffer, sizeof(buffer)))
		{
			// 0 = whole match
			// 1 = sub-match ()
			// 2 = sub-match ()
			if(regexclassname.Match(buffer) == 3)
			{
				IsClassnameEntry = true;

				ListClassnameOutputs.Rewind();

				regexclassname.GetSubString(1, NetClass, sizeof(NetClass), 0);
				regexclassname.GetSubString(2, buffer, sizeof(buffer), 0);

				ListClassnameOutputs.JumpToKey(buffer, true);
				ListClassnameOutputs.JumpToKey("keys", true);
				ListClassnameOutputs.GoBack();
				ListClassnameOutputs.JumpToKey("inputs", true);
				ListClassnameOutputs.GoBack();
				ListClassnameOutputs.JumpToKey("outputs", true);
				ListClassnameOutputs.GoBack();

				PrintToServer("classname: %s", buffer);
			}

			if(IsClassnameEntry)
			{
				if(regexoutput.Match(buffer) == 3)
				{
					regexoutput.GetSubString(1, skeyindex, sizeof(skeyindex), 0);
					regexoutput.GetSubString(2, buffer, sizeof(buffer), 0);
					Format(skeyindex, sizeof(skeyindex), "outputs/%s", skeyindex);
					ListClassnameOutputs.SetString(skeyindex, buffer);
					//PrintToServer("output: %s - %s", skeyindex, buffer);
				}
				else if(regexinput.Match(buffer) == 3)
				{
					regexinput.GetSubString(1, skeyindex, sizeof(skeyindex), 0);
					regexinput.GetSubString(2, buffer, sizeof(buffer), 0);
					Format(skeyindex, sizeof(skeyindex), "inputs/%s", skeyindex);
					ListClassnameOutputs.SetString(skeyindex, buffer);
					//PrintToServer("input: %s - %s", skeyindex, buffer);
				}
				else if(regexkey.Match(buffer) == 3)
				{
					fieldtype = PropField_Unsupported;

					regexkey.GetSubString(1, skeyindex, sizeof(skeyindex), 0);

					if(FindSendPropInfo(NetClass, skeyindex, fieldtype) != -1)
					{
						regexkey.GetSubString(2, buffer, sizeof(buffer), 0);
						Format(skeyindex, sizeof(skeyindex), "keys/%s (%s)", skeyindex, listfieldtypes[fieldtype]);
						ListClassnameOutputs.SetString(skeyindex, buffer);
						//PrintToServer("%s - %s", skeyindex, buffer);
					}
					else
					{
						regexkey.GetSubString(2, buffer, sizeof(buffer), 0);
						Format(skeyindex, sizeof(skeyindex), "keys/%s", skeyindex);
						ListClassnameOutputs.SetString(skeyindex, buffer);
						//PrintToServer("%s - %s", skeyindex, buffer);
					}
				}
			}
		}
		else
		{
			LogAction(-1, -1, "\nError in position %i / line %i\n", prop_data_file.Position, counter);

			//...ok, something happened. Lets also skip next keys,inputs,outputs and look next classname entry to continue read.
			IsClassnameEntry = false;
		}

		if(counter%1000 == 0)
		{
			NoteTimePassed(GetTime() - g_timestamp, counter);
		}
	}

	// Continue loop
	if(!IsEnd)
		RequestFrame(nextframe, prop_data_file);

}


void NoteTimePassed(int duration, int lines)
{
	PrintToServer("%i lines have read. Time has passed %02i:%02i:%02i",
										lines,
										(duration/3600)%24,
										(duration/60)%60,
										duration%60);
}

KeyValues MapOutputs;

public Action lookoutputs(int client, int args)
{

	char mapname[MAX_NAME_LENGTH];
	GetCurrentMap(mapname, sizeof(mapname));
	Format(mapname, sizeof(mapname), "%s_outputs.kv", mapname);

	if(MapOutputs != null)
		delete MapOutputs;

	MapOutputs = new KeyValues("data");

	int entity = -1;
	char clsname[65];
	char section[65];
	char value[65];
	char sindex[100];
	char targetname[65];


	while((entity = FindEntityByClassname(entity, "*")) != -1)
	{
		if(!IsValidEntity(entity) || !GetEntityClassname(entity, clsname, sizeof(clsname)))
			continue;

		//PrintToServer("\nclassname %s", clsname);

		ListClassnameOutputs.Rewind();
		MapOutputs.Rewind();

		if(!ListClassnameOutputs.JumpToKey(clsname))
		{
			continue;
		}

		Format(sindex, sizeof(sindex), "%i", entity);
		MapOutputs.JumpToKey(sindex, true);
		MapOutputs.SetString("classname", clsname);

		GetEntPropString(entity, Prop_Data, "m_iName", targetname, sizeof(targetname));

		if(!strlen(targetname))
		{
			Format(targetname, sizeof(targetname), " ");
		}

		MapOutputs.SetString("targetname", targetname);


		if(!ListClassnameOutputs.JumpToKey("outputs")
			|| !ListClassnameOutputs.GotoFirstSubKey(false))
		{
			continue;
		}


		do
		{
			ListClassnameOutputs.GetSectionName(section, sizeof(section));
			ListClassnameOutputs.GetString(section, value, sizeof(value));
			//PrintToServer("section %s", section);

			Checkoutput(entity, section, clsname, value);
		}
		while(ListClassnameOutputs.GotoNextKey(false))
	}


	MapOutputs.Rewind();
	MapOutputs.ExportToFile(mapname);
}



//public void Output_OnEntityOutput(const char[] output, int caller, int activator, float delay)
void Checkoutput(int caller, const char[] output, const char[] classname, const char[] outputname)
{
	//if (!IsValidEntity(caller))
	//{
	//	return;
	//}

	PrintToServer("\n%20s %i", classname, caller);

	char target[256];
	char input[256];
	char params[256];
	
	Address address = Address_Null;
	while ((address = FindEntityOutputAction(address, caller, output)) != Address_Null)
	{
		target[0] = '\0';
		input[0] = '\0';
		params[0] = '\0';

		GetEntityOutputActionTarget(address, target, sizeof(target));
		//PrintToServer("target %s", target);
		

		GetEntityOutputActionInput(address, input, sizeof(input));
		//PrintToServer("input %s", input);
		

		GetEntityOutputActionParams(address, params, sizeof(params));
		//PrintToServer("params %s", params);

		float delay = GetEntityOutputActionDelay(address);
		int timesToFire = GetEntityOutputActionTimesToFire(address);
		int IDStamp = GetEntityOutputActionIDStamp(address);

		if(strlen(target) || strlen(input) || strlen(params))
		{
			MapOutputs.SetString("outputname", outputname);
			MapOutputs.SetString("target", target);
			MapOutputs.SetString("input", input);
			MapOutputs.SetString("params", params);
			MapOutputs.SetFloat("delay", delay);
			MapOutputs.SetNum("once", timesToFire);
			MapOutputs.SetNum("id", IDStamp);

			PrintToServer("Output name: %s", outputname);
			PrintToServer("Target(s) name: %s", target);
			PrintToServer("Target Input: %s", input);
			PrintToServer("Parameters: %s", params);
			PrintToServer("Delay: %f, Only Once: %i, IDStamp %i", delay, timesToFire, IDStamp);
		}
	}
}